/*
 * Good Bad Attitude 2
 * by Javantea
 * June 2, 2006

Description:
A better version of Good Bad Attitude.
This grabs a list of good processes from /proc, then it monitors /proc and 
kills any new process. It is meant to be used in extremely hostile 
environments. It is a general use tool, but it can and should be modified as 
necessary.

It must be written in C, PHP, etc. The more the better for more attack vectors.
It must use the /proc interface and be as efficient as possible.
How do we know they use /proc?
How do we know that they use a standard ps?
How do we know...
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <signal.h>

// Way too many, but who cares.
#define INITIAL_GOOD_PROCS 16384
#define MAX_BAD_PROCS 32768
#define TOO_BIG_PROCS 60000
#define SLEEP_LENGTH_USEC 1000

/*
 * get_procs(): A process id list function.
 * Uses opendir() to find a list of processes in /proc.
 * returns the number of currently running processes and an array of process 
 * ids in procs.
 * Input: procs is an array that has been allocated already. length is the 
 * length of procs.
 */
int get_procs(int *procs, int length)
{
	int i = 0;
	DIR *proc_dir;
	proc_dir = opendir("/proc");
	
	struct dirent *d;
	while((d = readdir(proc_dir)) != 0)
	{
		// atoi doesn't detect whether it's non-zero, so we do.
		if(d->d_name[0] < '0' || d->d_name[0] > '9') continue;
		int proc_id = atoi(d->d_name);
		// Extra error checking.
		if(proc_id <= 0) continue;
		
		// Don't allow them to overflow the buffer.
		if(i >= length)
		{
			i++;
			if(i > TOO_BIG_PROCS)
			{
				printf("Error: Way too big process list: %i. Something must be wrong.\n", i);
				return i;
			}
			continue;
		}
		//printf("%s %i, ", d->d_name, proc_id); //d->d_type);
		procs[i] = proc_id;
		i++;
	}

	closedir(proc_dir);
	//printf("\n");
	
	return i;

}

#define MAX(a, b) ((a) > (b) ? (a) : (b))

/*
 * a_in_b(): An array search function.
 * returns the position in b of a. -1 otherwise.
 * Input: a is the search term, b is an array to search, 
 * b_len is the length of b.
 */
int a_in_b(int a, int *b, int b_len)
{
	int i;
	for(i = 0; i < b_len; i++)
	{
		if(a == b[i])
			return i;
	}
	return -1;
}

int main(int argc, char **argv)
{
	int *good_procs = (int *)malloc(INITIAL_GOOD_PROCS * sizeof(int));

	int good_proc_len = get_procs(good_procs, INITIAL_GOOD_PROCS);
	// If they have more, then try again with a massive buffer.
	if(good_proc_len > INITIAL_GOOD_PROCS)
	{
		printf("Warning: Increasing buffer size to %i.\n", good_proc_len);
		free(good_procs);
		good_procs = (int *)malloc(good_proc_len * sizeof(int));
		int good_proc_len2 = get_procs(good_procs, good_proc_len);
		if(good_proc_len2 > good_proc_len)
		{
			printf("Warning: larger buffer didn't help much: %i/%i.\n", good_proc_len2, good_proc_len);
		}
		
		printf("Good Processes\n");
		int i;
		for(i = 0; i < good_proc_len; i++)
		{
			printf("%i ", good_procs[i]);
		}
		printf("\n");
	}
	
	// At least twice as many as current or default.
	int bad_procs_size = MAX(good_proc_len * 2, MAX_BAD_PROCS);
	
	int *bad_procs = (int *)malloc(bad_procs_size * sizeof(int));
	
	while(1)
	{
		int bad_proc_len = get_procs(bad_procs, bad_procs_size);
		if(bad_proc_len > bad_procs_size)
		{
			printf("Warning: Too many bad procs: %i.\n", bad_proc_len);
			// Don't let them overflow our buffer.
			bad_proc_len = bad_procs_size;
		}
		int i;
		for(i = 0; i < bad_proc_len; i++)
		{
			// If it is not found in the previous list, kill it.
			if(a_in_b(bad_procs[i], good_procs, good_proc_len) < 0)
			{
				// tkill is the same as kill, but it only runs on single procs.
				int ret = kill(bad_procs[i], SIGKILL);
				printf("Killed %i: %i\n", bad_procs[i], ret);
			}
		}
		usleep(SLEEP_LENGTH_USEC);
	}
	free(bad_procs);
	free(good_procs);
	
	return 0;

}
